from dataclasses import dataclass
import numpy as np


@dataclass
class BulletDynamicsInfo:
    """
    Information pertinent to the dynamics of a rigid body link.
    More info at: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#

    """

    mass: float  # mass in kg
    lateral_friction: float  # friction coefficient
    local_inertia_diagonal: (
        np.ndarray
    )  # local inertia diagonal. Note that links and base are centered around the center of mass and aligned with the principal axes of inertia.
    local_inertial_pos: (
        np.ndarray
    )  # position of inertial frame in local coordinates of the joint frame
    local_inertial_orn: (
        np.ndarray
    )  # orientation of inertial frame in local coordinates of joint frame
    resolution: float  # coefficient of restitution
    rolling_friction: float  # rolling friction coefficient orthogonal to contact normal
    spinning_friction: float  # spinning friction coefficient around contact normal
    contact_damping: float  # -1 if not available. damping of contact constraints.
    contact_stiffness: float  # -1 if not available. stiffness of contact constraints.
    body_type: int  # 1=rigid body, 2 = multi body, 3 = soft body
    collision_margin: float  # advanced/internal/unsupported info. collision margin of the collision shape. collision margins depend on the shape type, it is not consistent.

    def __post_init__(self):
        self.local_inertia_diagonal = np.array(self.local_inertia_diagonal)
        self.local_inertial_pos = np.array(self.local_inertial_pos)
        self.local_inertial_orn = np.array(self.local_inertial_orn)


@dataclass
class BulletJointInfo:
    """
    Information pertinent to the joints of a rigid body linkage.
    More info at: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#

    """

    joint_index: int  # the same joint index as the input parameter
    joint_name: str  # the name of the joint, as specified in the URDF (or SDF etc) file
    joint_type: int  # type of the joint, this also implies the number of position and velocity variables. JOINT_REVOLUTE, JOINT_PRISMATIC, JOINT_SPHERICAL, JOINT_PLANAR, JOINT_FIXED. See the section on Base, Joint and Links for more details.
    q_index: (
        int  # the first position index in the positional state variables for this body
    )
    u_index: (
        int  # the first velocity index in the velocity state variables for this body
    )
    flags: int  # reserved
    joint_damping: float  # the joint damping value, as specified in the URDF file
    joint_friction: float  # the joint friction value, as specified in the URDF file
    joint_lower_limit: (
        np.ndarray
    )  # Positional lower limit for slider and revolute (hinge) joints, as specified in the URDF file.
    joint_upper_limit: (
        np.ndarray
    )  # Positional upper limit for slider and revolute joints, as specified in the URDF file. Values ignored in case upper limit <lower limit
    joint_max_force: float  # Maximum force specified in URDF (possibly other file formats) Note that this value is not automatically used. You can use maxForce in 'setJointMotorControl2'
    joint_max_velocity: float  # Maximum velocity specified in URDF. Note that the maximum velocity is not used in actual motor control commands at the moment.
    link_name: str  # the name of the link, as specified in the URDF (or SDF etc.) file
    joint_axis: np.ndarray  # joint axis in local frame (ignored for JOINT_FIXED)
    parent_frame_pos: np.ndarray  # joint position in parent frame
    parent_frame_orn: (
        np.ndarray
    )  # joint orientation in parent frame (quaternion x,y,z,w)
    parent_index: int  # parent link index, -1 for base

    def __post_init__(self):
        self.joint_lower_limit = np.array(self.joint_lower_limit)
        self.joint_upper_limit = np.array(self.joint_upper_limit)
        self.parent_frame_pos = np.array(self.parent_frame_pos)
        self.parent_frame_orn = np.array(self.parent_frame_orn)


@dataclass
class VisualShapeData:
    """
    Visual shape data fields. The return of p.getVisualShapeData(...)
    more info at: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#
    """

    obj_unique_id: int  # object unique id, same as the input
    link_id: int  # link index or -1 for the base
    vis_geom_type: int  # visual geometry type (TBD)
    dims: (
        np.ndarray
    )  # vec3, list of 3 floats, dimensions (size, local scale) of the geometry
    mesh_asset_fname: str  # path to the triangle mesh, if any. Typically relative to the URDF, SDF or MJCF file location, but could be absolute.
    local_vis_frame_pos: (
        np.ndarray
    )  # position of local visual frame, relative to link/joint frame
    local_vis_frame_orn: (
        np.ndarray
    )  # orientation of local visual frame relative to link/joint frame
    rgba_colour: np.ndarray  # URDF color (if any specified) in red/green/blue/alpha
    # texture_uid: int  # (field only exists if using VISUAL_SHAPE_DATA_TEXTURE_UNIQUE_IDS flags)Texture unique id of the shape, or -1 if none

    def __post_init__(self):
        self.dims = np.array(self.dims)
        self.local_vis_frame_pos = np.array(self.local_vis_frame_pos)
        self.local_vis_frame_orn = np.array(self.local_vis_frame_orn)
        self.rgba_colour = np.array(self.rgba_colour)


@dataclass
class BulletContactInfo:
    """
    Contact information between two rigid body links
    More info at: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#

    """

    contact_flag: int  # reserved
    body_uid_a: str  # body unique id of body A
    body_uid_b: int  # body unique id of body B
    link_id_a: int  # link index of body A, -1 for base
    link_id_b: int  # link index of body B, -1 for base
    position_on_a: np.ndarray  # contact 3D position on A, in Cartesian world coordinate
    position_on_b: (
        np.ndarray
    )  # contact 3D position on B, in Cartesian world coordinates
    contact_normal_on_b: np.ndarray  # contact normal on B, pointing towards A
    contact_distance: (
        float  # contact distance, positive for separation, negative for penetration
    )
    normal_force: float  # normal force applied during the last 'world.step()'
    lateral_friction1: (
        float  # lateral friction force in the lateralFrictionDir1 direction
    )
    lateral_friction_dir1: np.ndarray  # first lateral friction direction
    lateral_friction2: (
        float  # lateral friction force in the lateralFrictionDir2 direction
    )
    lateral_friction_dir2: np.ndarray  # second lateral friction direction

    def __post_init__(self):
        self.position_on_a = np.array(self.position_on_a)
        self.position_on_b = np.array(self.position_on_b)
        self.contact_normal_on_b = np.array(self.contact_normal_on_b)
        self.lateral_friction_dir1 = np.array(self.lateral_friction_dir1)
        self.lateral_friction_dir2 = np.array(self.lateral_friction_dir2)



from pathlib import Path
from tempfile import NamedTemporaryFile

from ament_index_python.packages import get_package_share_path
from urdf_parser_py.urdf import URDF, Mesh


def normalise_urdf_paths(file: Path | str) -> str:
    """Normalise the robot description file to use absolute paths for the collision and visual tags.

    Args:
        file (pathlib.Path | str): a path to a URDF, or pre-loaded URDF string
    Returns:
        (str): a URDF file path with absolute paths to collision and visual tag mesh resources

    """
    if isinstance(file, str):
        robot = URDF.from_xml_string(file)
    elif isinstance(file, Path):
        robot = URDF.from_xml_file(file)

    for link in robot.links:
        for collision in link.collisions:
            if isinstance(
                collision.geometry,
                Mesh,
            ) and collision.geometry.filename.startswith("package://"):
                package_name, relative_path = collision.geometry.filename.split(
                    "package://",
                )[1].split("/", 1)
                collision.geometry.filename = (
                    f"file:/{get_package_share_path(package_name)}/{relative_path}"
                )
        for visual in link.visuals:
            if isinstance(
                visual.geometry,
                Mesh,
            ) and visual.geometry.filename.startswith("package://"):
                package_name, relative_path = visual.geometry.filename.split(
                    "package://",
                )[1].split("/", 1)
                visual.geometry.filename = (
                    f"file:/{get_package_share_path(package_name)}/{relative_path}"
                )

    with NamedTemporaryFile(
        mode="w",
        prefix="grip_",
        delete=False,
    ) as parsed_file:
        parsed_file_path = parsed_file.name
        parsed_file.write(robot.to_xml_string())
        return parsed_file_path



import time
import pybullet as p
import os, copy
import pybullet_data as pd
from .bullet_robot import BulletRobot
from .gripper_interface import GripperInterface
from .end_effector_registry import EndEffectorRegistry

base_dir = pd.getDataPath()
PANDA_WITH_DEFAULT_GRIPPER_URDF = os.path.join(base_dir, "franka_panda/panda.urdf")


@EndEffectorRegistry.register("panda_hand")
class BulletPandaGripper(BulletRobot, GripperInterface):
    """
    The gripper interface for default panda gripper, which is a two-finger parallel gripper
    with two controllable joints.

    :param str urdf_file: urdf file for robot arm with default gripper
    :param int ee_index: reference link index for end effector
    :param [str] joint_names: list of joint names
    :param bool has_kinematics: to/not to extend arm with gripper kinematic chain
    :param int max_force: maximum finger force to apply
    """

    # Gripper type name for real gripper class (in arm farm layer)
    HW_INTERFACE = "arm_farm_interface.gripper_interface:DefaultFrankaGripper"

    def __init__(self, **kwargs):
        kwargs["urdf_file"] = kwargs.get("urdf_file", PANDA_WITH_DEFAULT_GRIPPER_URDF)
        kwargs["ee_index"] = kwargs.get("ee_index", -1)
        kwargs["tip_link"] = kwargs.get("tip_link", "panda_hand")

        kwargs["joint_names"] = ["panda_finger_joint1", "panda_finger_joint2"]

        kwargs["max_force"] = kwargs.get("max_force", 15)

        super().__init__(has_kinematics=False, **kwargs)

        self.physical_links = [8, 9, 10]
        self.home_positions = copy.deepcopy(self._ul)

        self.close_angles, self.open_angles = self.joint_limits

        self.reset_angles(self.home_positions)

    @property
    def ee_height(self):
        return self.ee_pose[0][2]

    def ee2ik_frame_pose(self, pos, quat):
        return pos, quat

    def apply_gripper_delta_action(self, action):
        """
        control gripper with a delta finger distance action.

        :param float action: delta joint angle action for all joints of the phase 4 gripper
        """
        positions = self.angles

        for i in range(len(self.joint_ids)):
            positions += action

            positions[i] = min(self._ul[i], max(self._ll[i], positions[i]))

            self.set_angles(positions, positionGain=1)

    def control_fingers(self, mode="close"):
        """
        A hard-coded opening/closing finger process with constant speed +/-0.05 per step time

        :param str mode: 'open' or 'close' mode
        """

        delta_action = 0.005
        n_update = 10

        if mode == "close":
            action = -delta_action
        else:
            action = delta_action

        for _ in range(n_update):
            self.apply_gripper_delta_action(action)
            p.stepSimulation(physicsClientId=self.phys_id)
            if mode == "close":
                time.sleep(3 / 240.0)

    def is_ready(self) -> bool:
        """Simulated gripper is always ready

        Returns:
            (bool): whether or not this gripper is ready (in sim it always is)
        """
        return True

    def validate_grasp(self) -> bool:
        """Returns true if the gripper is holding an object

        Returns:
            bool: The gripper has a valid grasp on the object
        """

        self.control_fingers(mode="close")
        return self.angles[0] > 0.01



import pybullet as p
import copy
import numpy as np
from ..io import *
from ..math import quat_multiply, rpy2quaternion, multiply_transform, invert_transform

from .robot_interface import RobotInterface
from .entity import Entity
from .world import BulletWorld
from .gui import BulletSliders
from .types import BulletDynamicsInfo, BulletJointInfo, VisualShapeData

from typing import Tuple, List, Dict, Optional


class BulletRobot(RobotInterface, Entity):
    """
    The BulletRobot represents an interface and connection to a robot
    An object of this class is exposes all generic functionalities to control a robot


    Args:
        urdf_file (str): full path to a robot URDF description
        id (int): if the robot entity already exists in the physics engine, then the id of the robot can be passed instead of its urdf_file
        parent (grip.robot.BulletRobot, optional): This robot may be a sub-part of an already existing parent robot (e.g. a robot hand attached to a robot arm)
        phys_id (int): physics engine identifier (the id of a BulletWorld)
        use_fixed_base (bool): Whether or not to make the object fixed in place. Default: True.
        position (numpy.ndarray): shapebase position of the robot
        orientation (numpy.ndarray): base orientation of the robot as en orientation quaterion following JPL format [x, y, z, w]
        ee_index (int, optional): end-effector link id. Defaults to None. Note: either ee_index or tip_link must be set, but not both.
        tip_link (str, optional): end-effector tip link name. Specifies the end-effector tip link, if ee_index is None, then ee_index = self.get_link_id(tip_link). Defaults to None.
        Note: either ee_index or tip_link must be set, but not both.
        ee_mount_link (str, optional): end-effector mount link name, used to specify up at which link the end-effector has been mounted on the robot.
        joint_names (str, optional): joint names that this robot instance should be controlling (can be a subset of the total number of joints of the robot)
            if not passed as a parameter, or None is given, then this BulletRobot will use all joint names as defined in the urdf description
        has_kinematics (bool): enable or disable forward and inverse kinematics with this robot (this should be usually true, but some robots may not need it)
        max_force (float): maximum generalised force
        sliders (bool): whether or not joint sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")
        ee_sliders (bool): whether or not cartesian end-effector sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")
    """

    def __init__(self, **kwargs):
        self.urdf_file = kwargs.get("urdf_file", None)
        self.data_path = kwargs.get("data_path", get_data_path())

        log.debug("Data path: {}".format(self.data_path))
        p.setAdditionalSearchPath(self.data_path)

        self.phys_id = kwargs.get("phys_id", None)
        self._useFixedBase = kwargs.get("use_fixed_base", True)
        self._position = np.array(kwargs.get("position", [0.0, 0.0, 0.0]))
        self._orientation = np.array(kwargs.get("orientation", [0, 0, 0, 1]))

        self.max_force = kwargs.get(
            "max_force", 500
        )  # max force when a joint in pos/vel controlled
        self.joint_friction = kwargs.get(
            "joint_friction", 0.01
        )  # friction force when a joint is in full torque mode

        self._sliders = kwargs.get("sliders", False)
        self._ee_sliders = kwargs.get("ee_sliders", False)

        # flags = p.URDF_USE_SELF_COLLISION#|p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
        # flags = p.URDF_USE_SELF_COLLISION #p.URDF_ENABLE_CACHED_GRAPHICS_SHAPES
        flags = 0
        self.parent = kwargs.get("parent", None)

        self.id = kwargs.get("id", None)

        self.texture_randomiser = kwargs.get("texture_randomiser", None)

        # assert self.parent is None or self.urdf_file is None and (not (self.parent is None and self.urdf_file is None)), \
        #     "Invalid instance. Provide either parent or URDF file, not both!"

        self.joint_idx0 = 0

        self.reset_home_positions = True
        if self.parent is None and self.id is None:
            self.id = p.loadURDF(
                self.urdf_file,
                self._position,
                self._orientation,
                useFixedBase=self._useFixedBase,
                physicsClientId=self.phys_id,
                flags=flags,
            )
        elif self.parent is not None and self.id is None:
            ## Else, this class is being attached to an existing body (e.g. gripper attached to arm)
            self.id = self.parent.id
            self.urdf_file = self.parent.urdf_file
            self.phys_id = self.parent.phys_id
            self.joint_idx0 = self.parent.n_joints
        else:
            # Else this is being attached to a robot existing somewhere else in the network
            self.reset_home_positions = False

        self._joint_names = kwargs.get("joint_names", None)

        self.ee_mount_link = kwargs.get("ee_mount_link", None)

        if self.ee_mount_link is not None and self._joint_names is None:
            _, self._joint_names = self.get_movable_joints(
                upto_link_name=self.ee_mount_link
            )

        self.joint_dict, self.link_dict = self.get_joint_dict()
        self.actuated_joint_dict = self.joint_dict.copy()
        self.link_dict["-1base"] = -1

        self.reverse_link_dict = dict([(v, k) for k, v in self.link_dict.items()])

        if self._joint_names is None:
            self._joint_names = list(self.joint_dict.keys())

        self.actuated_joint_names = self._joint_names.copy()
        self._setup_limits(**kwargs)

        if self.reset_home_positions:
            self.reset_angles(self.home_positions)

        # Kinematics setup
        self._kin = None
        has_kinematics_ = kwargs.get("has_kinematics", True)
        if has_kinematics_:
            self.setup_kinematics()

        # self.enable_torque_sensor(True)

        self.joint_sliders = None
        self.ee_sliders = None
        self.setup_gui()

        self._base_cid = None

    def _setup_limits(self, **kwargs) -> None:
        self.joint_ids = [self.joint_dict[name]["idx"] for name in self.joint_names]
        self.actuated_joint_ids = self.joint_ids.copy()
        self._ll = [self.joint_dict[name]["lower_limit"] for name in self.joint_names]
        self._ul = [self.joint_dict[name]["upper_limit"] for name in self.joint_names]
        self._jr = [self.joint_dict[name]["joint_range"] for name in self.joint_names]
        self._rp = [self.joint_dict[name]["rest_posture"] for name in self.joint_names]
        self._q_mean = [self.joint_dict[name]["q_mean"] for name in self.joint_names]

        self._n_joints = len(self.joint_ids)

        self._ndof = len(list(self.joint_dict.keys()))

        self.home_positions = kwargs.get("home_positions", self._rp)

        self.tip_link = kwargs.get("tip_link", None)

        self.ee_index = kwargs.get("ee_index", None)

        if self.tip_link is None and self.ee_index is not None:
            self.tip_link = self.get_link_name(self.ee_index)
        elif self.tip_link is not None and self.ee_index is None:
            self.ee_index = self.get_link_id(self.tip_link)
        elif self.tip_link is None and self.ee_index is None:
            self.ee_index = -1
            self.tip_link = self.get_link_name(self.ee_index)
        elif self.tip_link is not None and self.ee_index is not None:
            if self.get_link_id(self.tip_link) != self.ee_index:
                log.warning(
                    f"Tip link {self.tip_link} id does not match self.ee_index, but it should. Current ee_index is set to link name {self.get_link_name(self.ee_index)}. Will set self.ee_index = self.get_link_id(self.tip_link)."
                )

                self.ee_index = self.get_link_id(self.tip_link)

        log.debug("Tip link: {}".format(self.tip_link))
        log.debug("Joint names: {}".format(self.joint_names))
        log.debug("Joint IDs: {}".format(self.joint_ids))

    def setup_kinematics(self) -> None:
        """
        Sets up robot kinematics solver. This enable calls to forward_kinematics and inverse_kinematics.
        """

        if self._kin:
            log.info(
                "Kinematics was already set-up. Clearing up previous instance and re-initialising it."
            )

            self._kin._release()
            self._kin = None

        self._kin = RobotKinematics(parent=self)

    def has_kinematics(self) -> bool:
        """
        Checkes if this robot has a kinematics solver setup. The kinematics solver is a damped least squares differential inverse kinematics solver native from Bullet.

        Returns:
            (bool): whether or not this robot has a kinematics solver setup.

        """

        return hasattr(self, "_kin") and self._kin is not None

    def release_kinematics(self) -> None:
        """
        Releases default bullet kinematics solver for this robot.
        This method has no effect if this robot has no default kinematics solver or if its solver has been already released.
        """
        if self.has_kinematics():
            self._kin._release()
            self._kin = None

    def __del__(self):
        self.release_kinematics()

    def setup_gui(self) -> None:
        """
        Sets up debug GUI, adding joint sliders and end-effector joint sliders (if they were set to be shown in the constructor)
        """

        if self.joint_sliders is not None:
            self.joint_sliders.remove()
        if self.ee_sliders is not None:
            self.ee_sliders.remove()

        if self._sliders:
            self.joint_sliders = BulletSliders(
                self.phys_id, self.joint_names, self._ll, self._ul, self.angles
            )

        if self._ee_sliders:
            ee_names = ["ee_x", "ee_y", "ee_z", "ee_roll", "ee_pitch", "ee_yaw"]
            ee_pos, ee_ori = self.ee_pose
            ee_rpy = np.array(
                p.getEulerFromQuaternion(ee_ori, physicsClientId=self.phys_id)
            )

            min_pos = [c - 1.0 for c in ee_pos]
            max_pos = [c + 1.0 for c in ee_pos]
            min_ori = [-np.pi * 2] * 3
            max_ori = [np.pi * 2] * 3

            ee_pose = np.hstack([ee_pos, ee_rpy]).tolist()
            max_vals = max_pos + max_ori
            min_vals = min_pos + min_ori

            map_func = lambda values: (
                np.array(values[:3]),
                np.array(rpy2quaternion(values[3:])),
            )

            self.ee_sliders = BulletSliders(
                self.phys_id, ee_names, min_vals, max_vals, ee_pose, map_func=map_func
            )

    ## Getters

    def get_joint_state(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Gets the current joint state of the robot, which includes joint positions, velocities, reaction forces and joint efforts.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]): a tuple of 4 arrays with shape-(ndof,), each respectively being joint positions, velocities, reaction foces and joint efforts

        """
        joint_angles = []
        joint_velocities = []
        joint_reaction_forces = []
        joint_efforts = []

        for idx in self.joint_ids:
            joint_state = p.getJointState(self.id, idx, physicsClientId=self.phys_id)

            joint_angles.append(joint_state[0])

            joint_velocities.append(joint_state[1])

            joint_reaction_forces.append(joint_state[2])

            joint_efforts.append(joint_state[3])

        return (
            np.array(joint_angles),
            np.array(joint_velocities),
            np.array(joint_reaction_forces),
            np.array(joint_efforts),
        )

    def get_joint_dict(self) -> Dict[str, dict]:
        """
        Gets a dictionary of dictionaries containing information about the joints of this robot.

        Returns:
            (dict): joint info dictionary in the format of ``dict: {'joint_name': dict { 'q_index': [str, str], 'lower_limit': float, 'upper_limit': float, 'rest_posture': float, 'q_mean': float, 'joint_range': float, 'idx': int }}``.
        """
        # attribute_list = ['jointIndex', 'jointName', 'jointType',
        #                   'qIndex', 'uIndex', 'flags',
        #                   'jointDamping', 'jointFriction', 'jointLowerLimit',
        #                   'jointUpperLimit', 'jointMaxForce', 'jointMaxVelocity', 'linkName']

        joint_dict = {}
        link_dict = {}

        for i in range(p.getNumJoints(self.id, physicsClientId=self.phys_id)):
            joint_info = p.getJointInfo(self.id, i, physicsClientId=self.phys_id)

            link_name = joint_info[12].decode("utf-8")
            link_dict[link_name] = i

            is_movable = joint_info[2] != p.JOINT_FIXED

            if is_movable:
                q_index = joint_info[3]
                joint_name = joint_info[1].decode("utf-8")
                joint_type = joint_info[2]
                joint_ll = joint_info[8]
                joint_ul = joint_info[9]
                joint_max_force = joint_info[10]
                joint_max_vel = joint_info[11]

                rest_posture = 0.5 * (joint_ll + joint_ul)
                joint_range = joint_ul - joint_ll

                joint_dict[joint_name] = {
                    "q_index": q_index,
                    "lower_limit": joint_ll,
                    "upper_limit": joint_ul,
                    "max_force": joint_max_force,
                    "max_vel": joint_max_vel,
                    "rest_posture": rest_posture,
                    "q_mean": copy.deepcopy(rest_posture),
                    "joint_range": joint_range,
                    "idx": i,
                    "link_name": link_name,
                }

        return joint_dict, link_dict

    def get_movable_joints(
        self, upto_link_name: Optional[str] = None
    ) -> Tuple[np.ndarray, List[str]]:
        """
        Gets the movable joint ids and list of joint names.
        Args:

        Returns:
            (Tuple[numpy.ndarray, List[str]]): tuple containing the array with shape-(ndof,) with movable joint IDs, and list of movable joint names.

        """
        movable_joints = []
        joint_names = []

        for i in range(p.getNumJoints(self.id, physicsClientId=self.phys_id)):
            joint_info = p.getJointInfo(self.id, i, physicsClientId=self.phys_id)
            q_index = joint_info[3]
            joint_name = joint_info[1]
            joint_type = joint_info[2]
            joint_ll = joint_info[8]
            jint_ul = joint_info[9]
            link_name = joint_info[12].decode("utf-8")

            if joint_type != p.JOINT_FIXED:
                movable_joints.append(i)

                log.debug("Qidx {} CurrIdx {}".format(q_index, i))
                joint_names.append(joint_name.decode("utf-8"))

            if link_name == upto_link_name:
                break

        return np.array(movable_joints), joint_names

    def get_joint_limits(
        self,
    ) -> Tuple[List[float], List[float], List[float], List[float]]:
        """
        Gets a tuple containing the lower and upper limits, as well as joint rage (upper-lower) and rest posture ``(upper+lower)*0.5``

        Returns:
            (Tuple[List[float], List[float], List[float], List[float]]): lower and upper limits, joint range and rest posture.

        """
        lower_lim = np.zeros(self.n_joints)

        upper_lim = np.zeros(self.n_joints)

        rest_posture = np.zeros(self.n_joints)

        joint_range = np.zeros(self.n_joints)

        for k, idx in enumerate(self.joint_ids):
            lower_lim[k] = p.getJointInfo(self.id, idx, physicsClientId=self.phys_id)[8]

            upper_lim[k] = p.getJointInfo(self.id, idx, physicsClientId=self.phys_id)[9]

            rest_posture[k] = 0.5 * (lower_lim[k] + upper_lim[k])

            joint_range[k] = upper_lim[k] - lower_lim[k]

        # [dict([x]) for x in zip(['upper'] * a.n_cmd(), a._bullet_robot.get_joint_limits()['upper'])]
        # {'lower': lower_lim, 'upper': upper_lim, 'mean': rest_posture, 'range': joint_range}
        return (
            lower_lim.tolist(),
            upper_lim.tolist(),
            joint_range.tolist(),
            rest_posture.tolist(),
        )

    def get_joint_info(self, joint_id: int) -> BulletJointInfo:
        """
        Returns

        Args:
            joint_id (int): joint identifier.

        Returns:
           BulletJointInfo: bullet joint information of given joint_id

        """

        return BulletJointInfo(
            *p.getJointInfo(self.id, joint_id, physicsClientId=self.phys_id)
        )

    def get_base_pose(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the current base pose of this robot, as a position and quaternion.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a position shape-(3,) and unit quaternion orientation shape-(4,) representing the pose of this robot.
        """
        pos, ori = p.getBasePositionAndOrientation(
            self.id, physicsClientId=self.phys_id
        )

        return np.asarray(pos), np.asarray(ori)

    def get_base_velocity(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the current base velocity of this robot, as a linear and angular velocity in cartesian world coordinates.

        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a linear shape-(3,) and angular shape-(3,) velocities.
        """
        lin, ang = p.getBaseVelocity(self.id, physicsClientId=self.phys_id)

        return np.asarray(lin), np.asarray(ang)

    def get_link_ids(self, link_names: Optional[list[str]] = None) -> List[int]:
        """
        Gets the list of link IDs of this robot.

        Args:
            link_names (list[str], optional): optional list of link names. If not passed method returns all link_ids of this kinematic chain
        Returns:
            (List[int]): list of unique identifiers the links belonging to this robot, as loaded by the physics engine.
        """

        if link_names:
            # link ids for selected link names
            return [self.get_link_id(lname) for lname in link_names]

        return [lid for _, lid in self.link_dict.items()]

    def get_link_name(self, link_id: int) -> str:
        """
        Gets link name of given link_id of this robot.

        Returns:
            (str): the name associated with link_id
        """

        return self.reverse_link_dict[link_id]

    def get_link_names(self) -> List[str]:
        """
        Gets the list of link names of this robot.

        Returns:
            (List[str]): list of link names of this robot, as loaded by the physics engine.
        """

        return [lname for lname, _ in self.link_dict.items()]

    def get_link_pose(self, link_id: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the current pose of a given link identified by its link_id

        Args:
            link_id (int): link identifier.
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a position shape-(3,) and unit quaternion orientation shape-(4,) representing the pose of the chosen link.
        """

        if link_id < 0:
            return self.get_base_pose()

        link_state = p.getLinkState(
            self.id,
            link_id,
            computeForwardKinematics=True,
            physicsClientId=self.phys_id,
        )

        pos = np.asarray(link_state[4])
        ori = np.asarray(link_state[5])  # jpl convention

        return pos, ori

    def get_link_pose_by_name(self, link_name: str) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the current pose of a given link identified by its link name

        Args:
            link_id (str): link name.
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a position shape-(3,) and unit quaternion orientation shape-(4,) representing the pose of the chosen link.
        """

        link_id = self.link_dict[link_name]

        return self.get_link_pose(link_id)

    def get_dynamics_info(self, link_name) -> BulletDynamicsInfo:
        """
        Gets dynamics information of chosen link as per loaded from URDF definition

        Args:
            link_id (str): link name.
        Returns:
            (BulletDynamicsInfo): dynamics information for chose link
        """

        link_id = self.get_link_id(link_name)

        info = p.getDynamicsInfo(self.id, link_id, physicsClientId=self.phys_id)

        return BulletDynamicsInfo(*info)

    def get_link_id(self, link_name: str) -> int:
        """
        Gets the link ID from a given link name

        Args:
            link_id (str): link name.
        Returns:
            (int): link identifier
        """
        return self.link_dict[link_name]

    def get_link_velocity(self, link_id: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the velocity of given link identifier.

        Args:
            link_id (int): link identifier
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing two vectors of shape-(3,), respectively, linear and angular velocity vectors.
        """

        link_state = p.getLinkState(
            self.id, link_id, computeLinkVelocity=True, physicsClientId=self.phys_id
        )

        lin_vel = np.asarray(link_state[6])
        ang_vel = np.asarray(link_state[7])

        return lin_vel, ang_vel

    ## RobotInterface properties/getters

    @property
    def q_mean(self):
        """
        numpy.ndarray: mean joint angles, i.e. ``(upper+lower)*0.5``
        """
        return self._q_mean

    @property
    def n_joints(self):
        """
        int: total number of joints that this robot has.
        """
        return self._n_joints

    @property
    def angles(self):
        """
        numpy.ndarray: current joint angles/positions of this robot.
        """
        return self.get_joint_state()[0]

    def joint_angle(self, joint_idx: int) -> float:
        """
        Gets the current joint angle of a specified joint index.

        Args:
            joint_idx (int): joint index.
        Returns:
            (float): current joint angle/position for the specified joint index.
        """
        # [pos, vel, reac_forces, effort]
        joint_state = p.getJointState(self.id, joint_idx, physicsClientId=self.phys_id)

        state = [joint_state[0], joint_state[1], joint_state[3]]

        return state

    @property
    def joint_limits(self):
        """
        Tuple[List[float], List[float]]: pre-computed joint limits, as defined in URDF file.
        """
        return self._ll, self._ul

    @property
    def joint_names(self):
        """
        List[str]: list of joint names of this robot.
        """
        return self._joint_names

    @property
    def joint_velocities(self):
        """
        numpy.ndarray: current joint velocities of this robot.
        """
        return self.get_joint_state()[1]

    @property
    def joint_efforts(self):
        """
        numpy.ndarray: current joint efforts of this robot.
        """
        return self.get_joint_state()[3]

    @property
    def ee_pose(self):
        """
        Tuple[numpy.ndarray, numpy.ndarray]: position vector shape-(3,) and quaternion shape-(4,) representing the pose of this robot's end-effector.
        """
        return self.get_link_pose(self.ee_index)

    @property
    def ee_velocity(self):
        """
        Tuple[numpy.ndarray, numpy.ndarray]: tuple containing two vectors of shape-(3,), respectively, linear and angular velocity vectors of the end-effector.
        """
        return self.get_link_velocity(link_id=self.ee_index)

    @property
    def base_pose(self):
        """
        Tuple[numpy.ndarray, numpy.ndarray]: position vector shape-(3,) and quaternion shape-(4,) representing the pose of this robot base frame.
        """
        return self.get_base_pose()

    @property
    def base_velocity(self):
        """
        Tuple[numpy.ndarray, numpy.ndarray]: tuple containing two vectors of shape-(3,), respectively, linear and angular velocity vectors of the robot base.
        """
        return self.get_base_velocity()

    @property
    def kinematics(self) -> "grip.robot.RobotKinematics":
        """
        RobotKinematics: the default kinematics of this robot. May be None if has_kinematics=False at init.
        """
        return self._kin

    def state(self) -> Dict:
        """
        Returns the current state of this robot, which includes its joint names, positions, velocities, efforts, end-effector pose and a pointer to self.

        Returns:
            (dict): dictionary with the following format ``dict { 'joint_names': List[str], 'positions': numpy.ndarray, 'velocities': numpy.ndarray, 'efforts': numpy.ndarray, 'ee_pose': Tuple[numpy.ndarray, numpy.ndarray], 'entity': grip.robot.Entity}``
        """

        state = dict()

        state["joint_names"] = self.joint_names
        state["positions"] = self.angles
        state["velocities"] = self.joint_velocities
        state["efforts"] = self.joint_efforts
        state["ee_pose"] = self.ee_pose
        state["entity"] = self

        return state

    def reset(self) -> None:
        """
        Resets robot angles to its home position.
        """

        self.reset_angles(self.home_positions)
        self.set_angles(self.home_positions)

    def forward_kinematics(self, angles: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        If this robot has kinematics enabled, then this will return the forward kinematic solution for the given angles.

        Args:
            angles (numpy.ndarray): input joint angles.
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a position shape-(3,) and unit quaternion orientation shape-(4,) as the pose of the end-effector for this robot given the input angles.
        """
        assert self._kin, "Robot Kinematics has not been setup"

        position, orientation = self._kin.forward_kinematics(angles)

        return position, orientation

    def inverse_kinematics(self, position: np.ndarray, orientation: np.ndarray):
        """
        If this robot has kinematics enabled, inverse kinematic solution to reache the position and orientation given as input.

        Args:
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
        Returns:
            (numpy.ndarray): inverse kinematic joint positions solution for reaching the desired end-effector position and orientation
        """
        assert self._kin, "Robot Kinematics has not been setup"

        joint_angles = self._kin.inverse_kinematics(position, orientation)

        begin = self.joint_idx0
        end = self.joint_idx0 + self.n_joints
        return np.array(joint_angles)[begin:end]

    def accurate_inverse_kinematics(
        self,
        position: np.ndarray,
        orientation: np.ndarray,
        residual_threshold: float = 1e-4,
        max_iteration: int = 1000,
    ):
        """
        Solve ik to taret pose, return valid joint angles only when ik is accurate
        with residual value smaller than the specified threshold

        Args:
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
            residual_threshold (float): minimum translation residual value for ik to considered accurate
            max_iteration (int): maximum allowed iteration to the ik solving
        Returns:
            (numpy.ndarray or None): inverse kinematic joint positions solution for reaching the desired end-effector position and orientation. When no accurate solution is found, None is returned

        """
        n_iter = 0
        while n_iter < max_iteration:
            joint_angles = self.inverse_kinematics(position, orientation)
            ik_pos, ik_orn = self.forward_kinematics(joint_angles)
            tmp_residual = np.linalg.norm(position - ik_pos)
            close_enough = tmp_residual < residual_threshold
            if close_enough:
                return joint_angles

            n_iter += 1
        return None

    #### Methods with side effects below ###

    def reset_base_pose(self, position: np.ndarray, orientation: np.ndarray) -> None:
        """
        Sets the pose of the base-link of this robot.

        Args:
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
        """
        if self._base_cid is not None:
            p.changeConstraint(
                self._base_cid, position, orientation, maxForce=self.max_force
            )
        else:
            p.resetBasePositionAndOrientation(
                self.id, position, orientation, physicsClientId=self.phys_id
            )

        if self._kin:
            self._kin.set_base_pose(position, orientation)

        self._position = position
        self._orientation = orientation

    def set_link_pose(
        self, link_name: str, position: np.ndarray, orientation: np.ndarray
    ) -> None:
        """
        Sets a corresponding base pose such that desired link pose matches position and orientation.

        Args:
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
        """

        transform_wb_out = self.get_link_ref_base_pose(
            "-1base", link_name, position, orientation
        )

        self.reset_base_pose(*transform_wb_out)

    def get_link_ref_base_pose(
        self,
        base_link_name: str,
        link_name: str,
        position: np.ndarray,
        orientation: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets a corresponding base pose such that desired link pose matches position and orientation.

        Args:
            base_link_name (numpy.ndarray): base reference link name
            link_name (numpy.ndarray): desired link name
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
        Returns:
            (Tuple[np.ndarray, np.ndarray]): pose of link 'base_link_name' must have
                                            such that child link 'link_name' matches given position and orientation
        """

        # T_{base}^{world}
        transform_wb = self.get_link_pose_by_name(base_link_name)

        # T_{link}^{world}
        transform_wl = self.get_link_pose_by_name(link_name)

        # T_{world}^{link}
        transform_lw = invert_transform(*transform_wl)

        # T_{base}^{link}
        transform_lb = multiply_transform(
            transform_lw[0], transform_lw[1], transform_wb[0], transform_wb[1]
        )

        # Output: T_{world}^{base} (corresponding base pose such that link pose matches desired position and orientation)
        transform_wb_out = multiply_transform(
            position, orientation, transform_lb[0], transform_lb[1]
        )

        return transform_wb_out

    def reset_angles(self, angles: np.ndarray) -> None:
        """
        Resets the joint positions of this robot.

        Args:
            angles (numpy.ndarray): desired joint positions
        """
        for i, idx in enumerate(self.joint_ids):
            p.resetJointState(self.id, idx, angles[i], 0, physicsClientId=self.phys_id)

    def move_to_home(self) -> None:
        """
        Sends robot to its rest/home pose (rp)
        """
        self.set_angles(self.home_positions)

    def set_angles(
        self,
        angles: np.ndarray,
        joint_names: List[str] = None,
        positionGain: float = 0.03,
        velocityGain: float = 1,
    ) -> None:
        """
        Sends robot to desired joint angles using position control (internal Bullet PD Controller).

        Args:
            angles (numpy.ndarray): desired joint angles/positions shape-(N,), where N <= ndof. Typically you want to control all joints of the robot, so N=ndof.
            joint_names (List[str], optional): desired joint names to be controlled shape-(N,) (it could be a subset of the total number of joints (non mimic joints) of the robot). Important: has to match the shape of angles argument.
            positionGain (float): position gain (proportional gain of PD controller). Defaults to 0.03.
            velocityGain (float): derivative gain of the PD controller. Defaults to 1 (critically damped).
        """
        if len(angles) == 0:
            return

        joint_ids = self.actuated_joint_ids
        if joint_names is None:
            joint_names = self.actuated_joint_names
        else:
            joint_ids = [self.actuated_joint_dict[name]["idx"] for name in joint_names]

        n = len(joint_ids)

        # log.debug("Angle cmd: {}".format(angles))
        # log.debug("Joint IDs cmd: {}".format(joint_ids))
        # log.debug("Joint names cmd: {}".format(joint_names))
        # print("Angle cmd: {}".format(angles))
        # print("Joint IDs cmd: {}".format(joint_ids))
        # print("Joint names cmd: {}".format(joint_names))
        assert len(angles) == len(joint_names) and len(joint_ids) == len(
            angles
        ), "Number joint angle commands must match number of joint names"

        p.setJointMotorControlArray(
            bodyIndex=self.id,
            jointIndices=joint_ids,
            controlMode=p.POSITION_CONTROL,
            targetPositions=angles,
            targetVelocities=[0] * n,
            forces=[self.max_force] * n,  # 10000, 150
            positionGains=[positionGain] * n,  # 0.03,
            velocityGains=[velocityGain] * n,
            physicsClientId=self.phys_id,
        )

    def set_joint_velocities(self, cmd: np.ndarray, joint_names: List[str] = None):
        """
        Sends joint velocity control commands to the robot.

        Args:
            cmd (numpy.ndarray): desired joint velocities shape-(N,), where N <= ndof. Typically you want to control all joints of the robot, so N=ndof.
            joint_names (List[str], optional): desired joint names to be controlled shape-(N,) (it could be a subset of the total number of joints (non mimic joints) of the robot). Important: has to match the shape of cmd argument.
        """
        if len(cmd) == 0:
            return

        joint_ids = self.actuated_joint_ids
        n = len(joint_ids)

        if joint_names is None:
            joint_names = self.actuated_joint_names
        else:
            joint_ids = [self.actuated_joint_dict[name]["idx"] for name in joint_names]

        assert len(cmd) == len(joint_names) and len(joint_ids) == len(
            cmd
        ), "Number joint velocity commands must match number of joint names"

        p.setJointMotorControlArray(
            self.id,
            joint_ids,
            controlMode=p.VELOCITY_CONTROL,
            forces=[self.max_force] * n,
            targetVelocities=cmd,
            physicsClientId=self.phys_id,
        )

    def set_joint_torques(self, cmd: np.ndarray, joint_names: List[str] = None) -> None:
        """
        Sends joint torque control commands to the robot.

        Args:
            cmd (numpy.ndarray): desired joint velocities shape-(N,), where N <= ndof. Typically you want to control all joints of the robot, so N=ndof.
            joint_names (List[str], optional): desired joint names to be controlled shape-(N,) (it could be a subset of the total number of joints (non mimic joints) of the robot). Important: has to match the shape of cmd argument.
        """

        if len(cmd) == 0:
            return

        joint_ids = self.actuated_joint_ids
        if joint_names is None:
            joint_names = self.actuated_joint_names
        else:
            joint_ids = [self.actuated_joint_dict[name]["idx"] for name in joint_names]

        assert len(cmd) == len(joint_names) and len(joint_ids) == len(
            cmd
        ), "Number joint torque commands must match number of joint names given"

        p.setJointMotorControlArray(
            self.id,
            joint_ids,
            controlMode=p.TORQUE_CONTROL,
            forces=cmd,
            physicsClientId=self.phys_id,
        )

    def enable_torque_mode(
        self, joint_names: List[str] = None, enable: bool = True
    ) -> None:
        """
        By default, each revolute joint and prismatic joint is motorized using a velocity motor. You can disable those default motor by using a maximum force of 0. This will let you perform torque control.
        Note: If torque control is enabled, it is better to call stepsimulation in the control loop, and disable realtime sim of the pybullet world.

        Args:
            joint_names (List[str], optional): desired joint names to be controlled shape-(N,) (it could be a subset of the total number of joints of the robot). Important: has to match the shape of cmd argument.
            enable (bool): whether to enable or disable torque control for the specified joints.
        """
        if joint_names is None:
            joint_ids = self.actuated_joint_ids
        else:
            joint_ids = [self.actuated_joint_dict[name]["idx"] for name in joint_names]

        if enable:
            p.setJointMotorControlArray(
                self.id,
                joint_ids,
                controlMode=p.VELOCITY_CONTROL,
                forces=np.ones(len(joint_ids)) * self.joint_friction,
            )
        else:
            p.setJointMotorControlArray(
                self.id,
                joint_ids,
                controlMode=p.VELOCITY_CONTROL,
                forces=np.ones(len(joint_ids)) * self.max_force,
            )

    def set_ee_pose(self, pos: np.ndarray, quat: np.ndarray) -> None:
        """
        Cartesian position control. Sends robot's end-effector to desired cartesian pose (position, quaternion) using position control.

        Args:
            pos (numpy.ndarray): desired end-effector position.
            quat (numpy.ndarray): desired end-effector orientation.
        """
        joint_angles = self.inverse_kinematics(pos, quat)

        self.set_angles(joint_angles)

    def exec_cartesian_cmd(self, cmd_p: np.ndarray, cmd_q: np.ndarray) -> None:
        """
        Cartesian position control. Sends robot's end-effector to desired cartesian pose (position, quaternion) using position control.

        Args:
            cmd_p (numpy.ndarray): desired end-effector position.
            cmd_q (numpy.ndarray): desired end-effector orientation.
        """

        self.set_ee_pose(cmd_p, cmd_q)

    def exec_cartesian_delta_cmd(self, cmd_dp: np.ndarray, cmd_dq: np.ndarray) -> None:
        """
        Cartesian position control. Sends robot's end-effector to desired cartesian displacement pose.

        Args:
            cmd_dp (numpy.ndarray): desired end-effector position displacement.
            cmd_dq (numpy.ndarray): desired end-effector orientation displacement.
        """
        ee_p, ee_q = self.ee_pose

        cmd_p = np.add(ee_p, cmd_dp)
        cmd_q = quat_multiply(ee_q, cmd_dq)
        log.info("cmd_p: {} cmd_q {}".format(cmd_dp, cmd_dq))

        # log.info("Close {}".format(np.allclose(cmd_p,ee_p)))
        self.set_ee_pose(cmd_p, cmd_q)

    def exec_position_cmd(self, cmd: np.ndarray) -> None:
        """
        Sends robot to desired joint angles using position control (internal Bullet PD Controller).

        Args:
            cmd (numpy.ndarray): desired joint angles/positions shape-(ndof,).
        """
        self.set_angles(cmd)

    def exec_position_cmd_delta(self, cmd: np.ndarray) -> np.ndarray:
        """
        Sends robot to desired joint angles using position control displacement command (internal Bullet PD Controller).

        Args:
            cmd (numpy.ndarray): desired joint relative displacement shape-(ndof,).
        """
        angles = self.angles

        position_cmd = np.add(angles, cmd)

        self.set_angles(position_cmd)

    def exec_velocity_cmd(self, cmd: np.ndarray) -> None:
        """
        Sends robot to desired joint velocity commands to the robot.

        Args:
            cmd (numpy.ndarray): desired joint velocities shape-(ndof,).
        """
        self.set_joint_velocities(cmd)

    def exec_torque_cmd(self, cmd: np.ndarray) -> None:
        """
        Sends robot to desired joint torque commands to the robot.

        Args:
            cmd (numpy.ndarray): desired joint torques shape-(ndof,).
        """
        self.set_joint_torques(cmd)

    def jacobian(self, joint_angles: np.ndarray = None) -> np.ndarray:
        """
        Gets jacobian of the robot at specified joint_angles, or current if joint_angles=None.

        Args:
            joint_angles (numpy.ndarray, optional): if different than None, retrieves jacobian matrix at specified joint_angles array with shape-(ndof,).
        Returns:
            (numpy.ndarray): manipulator jacobian matrix shape(6, ndof).
        """

        begin = self.joint_idx0
        end = self.joint_idx0 + self.n_joints

        if joint_angles is None:
            joint_angles = self.angles
            # log.info("Jacobian::joint_angles {}".format(joint_angles))

        n = len(joint_angles)

        positions = joint_angles
        if n < self._ndof:
            positions = np.zeros(self._ndof)
            # log.info("Jacobian::positions {}".format(positions))
            # log.info("Jacobian::joint_ids: {}".format(self.joint_ids))
            positions[begin:end] = joint_angles

        lin_jac, ang_jac = p.calculateJacobian(
            bodyUniqueId=self.id,
            linkIndex=self.ee_index,
            localPosition=[0.0, 0.0, 0.0],
            objPositions=list(positions),
            objVelocities=[0] * self._ndof,
            objAccelerations=[0] * self._ndof,
            physicsClientId=self.phys_id,
        )

        lin_jac = np.array(lin_jac)
        ang_jac = np.array(ang_jac)

        jac = np.vstack([lin_jac, ang_jac])[:, begin:end]

        return jac

    def inertia(self, joint_angles: np.ndarray = None) -> np.ndarray:
        """
        Args:
            joint_angles (numpy.ndarray, optional): optional parameter, if None, then returned inertia is evaluated at current joint_angles. Otherwise, returned inertia tensor is evaluated at current joint angles.

        Return:
            (numpy.ndarray): Joint space inertia tensor
        """

        begin = self.joint_idx0
        end = self.joint_idx0 + self.n_joints

        if joint_angles is None:
            joint_angles = self.angles

        inertia_tensor = np.array(
            p.calculateMassMatrix(
                self.id, list(joint_angles), physicsClientId=self.phys_id
            )
        )

        return inertia_tensor[begin:end, begin:end]

    def random_texture(self) -> None:
        """
        If texture randomiser is present, randomises this robot's texture. Useful for domain randomisation applications.
        """
        if self.texture_randomiser is not None:
            self.texture_randomiser.randomise(self)

    def enable_torque_sensor(self, enable: bool, joint_id: int = None) -> None:
        """
        Enable or disable torque_sensor.
        """
        joint = self.actuated_joint_ids[-1] if joint_id is None else joint_id
        p.enableJointForceTorqueSensor(
            self.id, joint, enable, physicsClientId=self.phys_id
        )

    def get_contact_force(
        self, target_links: list, ignore_self_contact: bool = True
    ) -> np.ndarray:
        """
        Calculate contact forces applied onto given link(s)

        Args:
            target_links (int): list of target link indexes to inspect contact
            ignore_self_contact (bool): if self contact should be countered
        Returns:
            (numpy.ndarray): total contact force on target links
        """
        contact_force = np.array([0.0, 0.0, 0.0])
        all_contacts = self.contact_points()
        for contact in all_contacts:
            (
                unused_flag,
                body_a_id,
                body_b_id,
                link_a_id,
                unused_link_b_id,
                unused_pos_on_a,
                unused_pos_on_b,
                contact_normal_b_to_a,
                unused_distance,
                normal_force,
                friction_1,
                friction_direction_1,
                friction_2,
                friction_direction_2,
            ) = contact
            if (ignore_self_contact) and (body_b_id == body_a_id):
                continue  # Ignore self contacts
            if link_a_id in target_links:
                tmp_contact_force = np.array(contact_normal_b_to_a) * normal_force
                contact_force += tmp_contact_force
        return contact_force

    def get_friction_force(self, target_links: list) -> None:
        """
        Calculate lateral friction forces applied onto given link(s)

        Args:
            target_links (int): list of target link indexes to inspect lateral friction
        Returns:
            (numpy.ndarray): total lateral friction force on target_links
        """

        friction_force = np.array([0.0, 0.0, 0.0])
        all_contacts = self.contact_points()
        for contact in all_contacts:
            (
                unused_flag,
                body_a_id,
                body_b_id,
                link_a_id,
                unused_link_b_id,
                unused_pos_on_a,
                unused_pos_on_b,
                contact_normal_b_to_a,
                unused_distance,
                normal_force,
                friction_1,
                friction_direction_1,
                friction_2,
                friction_direction_2,
            ) = contact
            if body_b_id == body_a_id:
                continue  # Ignore self contacts
            elif link_a_id in target_links:
                tmp_friction_force = (
                    np.array(friction_direction_1) * friction_1
                    + np.array(friction_direction_2) * friction_2
                )
                friction_force += tmp_friction_force
        return friction_force

    def attach_object(self, link_name: str, entity: Entity) -> None:
        """
        Attaches entity to selected link of this robot.

        Args:
            link_name (str): link name of this robot to attach entity to.
            entity (grip.robot.Entity): entity to attach to specified link.
        """
        object_pos, object_ori = entity.pose
        link_pos, link_ori = self.get_link_pose_by_name(link_name)

        ilp, ilq = invert_transform(link_pos, link_ori)

        rp, rq = multiply_transform(ilp, ilq, object_pos, object_ori)

        entity.create_constraint(rp, rq, self, link_name)

    def detach_object(self, entity: Entity) -> None:
        """
        Detaches entity from this robot (assuming it has been attached before via self.attach_object method)
        """
        entity.remove_constraint()

    def disable_collision(self) -> None:
        """
        Disable collisions between this robot and everything else.
        """
        p.setCollisionFilterGroupMask(self.id, -1, 0, 0, physicsClientId=self.phys_id)
        link_ids = self.get_link_ids()

        for lid in link_ids:
            p.setCollisionFilterGroupMask(
                self.id, lid, 0, 0, physicsClientId=self.phys_id
            )

    def set_collision_filter(self, entity: Entity, enable: bool):
        link_ids = self.get_link_ids()

        for lid in link_ids:
            p.setCollisionFilterPair(
                self.id, entity.id, lid, -1, int(enable), physicsClientId=self.phys_id
            )

            if isinstance(entity, BulletRobot):
                entity_link_ids = entity.get_link_ids()
                for elid in entity_link_ids:
                    p.setCollisionFilterPair(
                        self.id,
                        entity.id,
                        lid,
                        elid,
                        int(enable),
                        physicsClientId=self.phys_id,
                    )

    def contact_points(self):
        """
        Returns contact points between this robot and anything that it may be contacting

        Returns:
            contact points: [bullet_contacts]
                returns list of bullet contact points
                see: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#heading=h.cb0co8y2vuvc

        """

        contact_points = p.getContactPoints(self.id, physicsClientId=self.phys_id)

        return contact_points

    def aabb(
        self, link_id: int = -1, expansion: float = 0.0
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets axis-aligned bounding box of a given link of this robot.
        Args:
            link_id (int, optional): link id of this multi-body. Defaults to -1, meaning base link.
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): a tuple of 3-d vectors (aabb_min, aabb_max) representing the minimum and maximum bounds in world frame.
        """
        assert link_id == -1 or link_id in self.get_link_ids()

        aabb = p.getAABB(self.id, link_id, physicsClientId=self.phys_id)

        aabb_min = aabb[0]
        aabb_max = aabb[1]

        return np.array(aabb_min) - expansion, np.array(aabb_max) + expansion

    def create_constraint(
        self,
        position: np.ndarray = np.zeros(3),
        orientation: np.ndarray = np.array([0.0, 0.0, 0.0, 1.0]),
        max_force: float = 200,
    ) -> None:
        """
        Creates a rigid body constraint for the base frame of this body.

        Args:
            position (numpy.ndarray): the orientation of the joint frame relative to parent center of mass coordinate frame
            orientation (numpy.ndarray): the orientation of the joint frame relative to the world origin coordinate frame
        """
        if self._base_cid is not None:
            return

        self._position_des = position
        self._orientation_des = orientation
        self._base_cid = p.createConstraint(
            self.id,
            -1,
            -1,
            -1,
            p.JOINT_FIXED,
            [0, 0, 0],
            [0, 0, 0],
            position,  # [0.500000, 0.300006, 0.700000],
            self._orientation_des,
            physicsClientId=self.phys_id,
        )

        # Note: the mysterious `erp` is of EXTREME importance
        p.changeConstraint(
            self._base_cid, maxForce=max_force, erp=0.8, physicsClientId=self.phys_id
        )

    def remove_constraint(self) -> None:
        """Removes constraint created with self.create_constraint"""
        if self._base_cid is None:
            return

        p.removeConstraint(self._base_cid, physicsClientId=self.phys_id)

        self._base_cid = None

    def setup_mimic_joints(
        self,
        mimic_parent_joint_name: str,
        mimic_children: List[str],
        maxForce: float = 100,
    ) -> None:
        """Sets up mimic joints by creating gear joint constraints between parent and children joints

        Args:
            mimic_parent_joint_name (str): parent joint name to be mimicked
            mimic_children (List[str]): joint names of child joints that will be mimicking the desired parent joint.
        """

        if mimic_parent_joint_name not in self.joint_dict:
            log.warning(
                f"Impossible to setup mimic joints - parent joint {mimic_parent_joint_name} not found!"
            )
            return
        self.mimic_parent_id = self.joint_dict[mimic_parent_joint_name]["idx"]
        self.id2mimic_multiplier = {}
        # Preliminary check: all mimic_children joints needs to be available to be coupled (no action must be taken bofore it)
        for joint_name in mimic_children:
            if joint_name not in self.actuated_joint_dict:
                log.warning(
                    f"Impossible to setup mimic joints - child joint {joint_name} not found or already coupled with another parent joint!"
                )
                return
        for joint_name in mimic_children:
            joint_id = self.actuated_joint_dict[joint_name]["idx"]
            mimic_multiplier = mimic_children[joint_name]
            self.id2mimic_multiplier[joint_id] = mimic_multiplier
            log.debug("Removing '{}' from actuated joint list".format(joint_name))
            self.actuated_joint_names.remove(joint_name)
            self.actuated_joint_ids.remove(joint_id)
            del self.actuated_joint_dict[joint_name]

        for joint_id, multiplier in self.id2mimic_multiplier.items():
            c = p.createConstraint(
                self.id,
                self.mimic_parent_id,
                self.id,
                joint_id,
                jointType=p.JOINT_GEAR,
                jointAxis=[0, 1, 0],
                parentFramePosition=[0, 0, 0],
                childFramePosition=[0, 0, 0],
                physicsClientId=self.phys_id,
            )
            # Note: the mysterious `erp` is of EXTREME importance
            p.changeConstraint(
                c,
                gearRatio=-multiplier,
                maxForce=maxForce,
                erp=1.0,
                physicsClientId=self.phys_id,
            )

        log.debug(
            "Actuated joint names (updated): {}".format(self.actuated_joint_names)
        )
        log.debug("Actuated joint IDs (updated): {}".format(self.actuated_joint_ids))

    def set_colour(self, rgba_colour: np.ndarray) -> None:
        """
        Sets robot colour.

        Args:
            rgba_colour: color components for RED, GREEN, BLUE and ALPHA, each in range [0..1]. Alpha has to be 0 (invisible) or 1 (visible) at the moment. Note that TinyRenderer doesn't support transparancy, but the GUI/EGL OpenGL3 renderer does.
        """
        shape_data = [
            VisualShapeData(*shape)
            for shape in p.getVisualShapeData(self.id, physicsClientId=self.phys_id)
        ]

        for shape in shape_data:
            p.changeVisualShape(
                self.id,
                shape.link_id,
                rgbaColor=rgba_colour,
                physicsClientId=self.phys_id,
            )

    def set_link_colour(self, rgba_colour: np.ndarray, link_ids: list[int]) -> None:
        """
        Sets robot link colour.

        Args:
            rgba_colour: color components for RED, GREEN, BLUE and ALPHA, each in range [0..1]. Alpha has to be 0 (invisible) or 1 (visible) at the moment. Note that TinyRenderer doesn't support transparancy, but the GUI/EGL OpenGL3 renderer does.
            link_ids: link indices to set desired colour
        """

        shape_link_ids = [
            VisualShapeData(*shape).link_id
            for shape in p.getVisualShapeData(self.id, physicsClientId=self.phys_id)
        ]

        assert all(
            lid in shape_link_ids for lid in link_ids
        ), "Link ids must be valid and exist/belong to this robot."

        for lid in link_ids:
            p.changeVisualShape(
                self.id, lid, rgbaColor=rgba_colour, physicsClientId=self.phys_id
            )


class RobotKinematics(object):
    def __init__(self, **kwargs):
        self.world = BulletWorld(phys_opt="direct")

        self._parent = kwargs.get("parent", None)

        log.debug("Parent: {}".format(self._parent.__dict__))
        log.debug("kinematics URDF: {}".format(self._parent.urdf_file))

        self.robot_arm = BulletRobot(
            urdf_file=self._parent.urdf_file,
            ee_index=self._parent.ee_index,
            joint_names=self._parent.joint_names,
            phys_id=self.world.id,
            position=self._parent._position,
            orientation=self._parent._orientation,
            home_positions=self._parent.home_positions,
            has_kinematics=False,
        )

    def _release(self):
        if self.world is not None and self.world.is_connected():
            self.world.disconnect()

        self.world = None

    def __del__(self):
        self._release()

    def set_base_pose(self, position, orientation):
        self.robot_arm.reset_base_pose(position, orientation)

    def forward_kinematics(self, angles) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the forward kinematic solution for the given angles.

        Args:
            angles (numpy.ndarray): input joint angles.
        Returns:
            (Tuple[numpy.ndarray, numpy.ndarray]): tuple containing a position shape-(3,) and unit quaternion orientation shape-(4,) as the pose of the end-effector for this robot given the input angles.
        """
        self.robot_arm.reset_angles(angles)

        return self.robot_arm.ee_pose

    def inverse_kinematics(self, position, orientation, seed=None) -> np.ndarray:
        """
        Computes and returns inverse kinematic solution to reach the position and orientation given as input.

        Args:
            position (numpy.ndarray): desired position array shape(3,).
            orientation (numpy.ndarray): desired orientation quaternion array shape(4,).
            seed (numpy.ndarray): Optional null-space IK rest poses.
        Returns:
            (numpy.ndarray): inverse kinematic joint positions solution for reaching the desired end-effector position and orientation
        """
        self.robot_arm.reset_angles(self._parent.angles)

        if seed is None:
            seed = self.robot_arm.angles  # self.robot_arm._rp

        joint_angles = p.calculateInverseKinematics(
            self.robot_arm.id,
            self.robot_arm.ee_index,
            position,
            orientation,
            self.robot_arm._ll,
            self.robot_arm._ul,
            self.robot_arm._jr,
            seed,
            jointDamping=[0.001] * self.robot_arm._ndof,
            maxNumIterations=1000,
            residualThreshold=1e-9,
            physicsClientId=self.world.id,
        )

        return np.array(joint_angles)



import time


class Timer:
    def __init__(self):
        self.t0 = time.time()
        self.duration = 0

    def start(self, duration=0):
        self.t0 = time.time()
        self.duration = duration

    def elapsed(self):
        return time.time() - self.t0

    def finished(self):
        return self.elapsed() >= self.duration



from typing import Union
from .joint_state_tools import JointStatePub, JointStateSub, LocalJointStatePub
from .trajectory_action_server import TrajectoryActionServer
from .trajectory_action_client import TrajectoryActionClient
from control_msgs.action import FollowJointTrajectory
import time

from ..robot import BulletRobotArm
from ..motion import Path
from .io import is_sim


class ROSRobotArm(BulletRobotArm):
    """A ROS robot that specialises a BulletRobotArm representing a robot arm which can optionally have a gripper as an end-effector. It exposes standard ROS interfaces mimicking and exposing the same basic topics and actions of a real robot. It is also able to also connect to a real robot.
    Represents a simulated or a real robot arm. It dynamically detects whether or not it should simulate or simply connect to a robot if a robot already exists.
    When connected to a real robot, the simulated robot will act as a slave, mirroing the state of the real robot.
    """

    def __init__(
        self, parent_node: "rclpy.node.Node", auto_sim=True, sim=True, **kwargs
    ):
        """ROS Robot Arm constructor

        Args:
            parent_node (rclpy.node.Node): associated parent node.
            gripper_type (str, optional): gripper type name from grip.robot.EndEffectorRegistry to be used as the robot end-effector. Defaults to None, meaning no gripper is created.
            gripper_max_force (float, optional): maximum gripper force. Defaults to 40N.
            urdf_file (str): full path to a robot URDF description
            id (int): if the robot entity already exists in the physics engine, then the id of the robot can be passed instead of its urdf_file
            parent (grip.robot.BulletRobot, optional): This robot may be a sub-part of an already existing parent robot (e.g. a robot hand attached to a robot arm)
            phys_id (int): physics engine identifier (the id of a BulletWorld)
            use_fixed_base (bool): Whether or not to make the object fixed in place. Default: True.
            position (numpy.ndarray): shapebase position of the robot
            orientation (numpy.ndarray): base orientation of the robot as en orientation quaterion following JPL format [x, y, z, w]
            ee_index (int, optional): end-effector link id. Defaults to None. Note: either ee_index or tip_link must be set, but not both.
            tip_link (str, optional): end-effector tip link name. Specifies the end-effector tip link, if ee_index is None, then ee_index = self.get_link_id(tip_link). Defaults to None.
            Note: either ee_index or tip_link must be set, but not both.
            ee_mount_link (str, optional): end-effector mount link name, used to specify up at which link the end-effector has been mounted on the robot.
            joint_names (str, optional): joint names that this robot instance should be controlling (can be a subset of the total number of joints of the robot)
            if not passed as a parameter, or None is given, then this BulletRobot will use all joint names as defined in the urdf description
            has_kinematics (bool): enable or disable forward and inverse kinematics with this robot (this should be usually true, but some robots may not need it)
            max_force (float): maximum generalised force
            sliders (bool): whether or not joint sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")
            ee_sliders (bool): whether or not cartesian end-effector sliders should be added to the debug GUI (only used if grip.robot.BulletWorld was created with phys_opt="gui")
            auto_sim (bool, optional): automatically detects if a real robot exists in the network and connects to it if true, this will yield self.sim=False the argument sim is ignored. Defaults True.
            sim (bool, optional): when auto_sim=False, choose whether this instance will be a simulation instance of a connection to the real robot which will mirror the robot state.
            enable_action_server (bool, optional): whether or not internal joint trajectory action server will be enabled.
        """
        super().__init__(**kwargs)

        self.node = parent_node

        self.sim = is_sim(self.node) if auto_sim else sim

        self._last_joint_command = None

        self.state_topic = kwargs.setdefault("state_topic", "/joint_states")
        self.command_topic = kwargs.setdefault("command_topic", "/joint_commands")
        self.enable_action_server = kwargs.setdefault("enable_action_server", True)
        self.enable_pub = kwargs.setdefault("enable_pub", True)

        if self.sim:
            # If we are in simulation we need to simulate the joint state publisher and the trajectory action server.

            if self.enable_pub:
                self.joint_state_pub = JointStatePub(
                    self.node,
                    self,
                    topic_name=self.state_topic,
                    command_topic_name=self.command_topic,
                )

                if self.gripper:
                    self.gripper_joint_state_pub = LocalJointStatePub(
                        self.node,
                        self.gripper,
                    )

                if self.enable_action_server:
                    self.trajectory_action_server = TrajectoryActionServer(
                        self.node, self
                    )

        else:
            self.joint_state_sub = JointStateSub(
                self.node,
                topic_name=self.state_topic,
                joint_update_callback=self.set_angles,
            )
            self.trajectory_client = TrajectoryActionClient(self.node)

    def start_async_ros_comm(self, rate: int = 10):
        self.joint_state_pub.setup_timer_publisher(rate)

        if self.gripper:
            self.gripper_joint_state_pub.setup_timer_publisher(rate)

    def follow_joint_path(
        self, trajectory: Union[Path, FollowJointTrajectory.Goal], dt: float = 0.01
    ):
        if self.sim:
            if isinstance(trajectory, FollowJointTrajectory.Goal):
                trajectory = Path.from_ros_trajectory_action_goal(trajectory)

            for wpt in trajectory:
                self.set_angles(wpt, joint_names=trajectory.joint_names)
                time.sleep(dt)
        else:
            if isinstance(trajectory, Path):
                trajectory = trajectory.as_ros_trajectory_action_goal()

            self.trajectory_client.send_trajectory_goal(trajectory)



import abc


class Sensor(abc.ABC):
    @abc.abstractmethod
    def obs(self):
        """
        Retrieves an observation of this sensor
        """
        raise NotImplementedError("obs: NO EFFECT, NOT IMPLEMENTED")



from typing import Tuple, List, Optional
import numpy as np
from tf2_ros import TransformBroadcaster
from geometry_msgs.msg import TransformStamped

from tf2_ros.buffer import Buffer
from tf2_ros.transform_listener import TransformListener
from rclpy.duration import Duration


class TFPublisher:
    """A transform publisher utility class. It can be used to broadcast transform to ROS TF tree."""

    def __init__(self, parent_node: "rclpy.node.Node"):
        """Contruscts a TFPublisher object

        Args:
            parent_node (rclpy.node.Node): parent node to which this TFPublisher is associated to.
        """

        self.node = parent_node

        self.tf_broadcaster = TransformBroadcaster(self.node)

    def broadcast_frame(
        self,
        transform: Tuple[List[float], List[float]],
        source_frame: Optional[str] = "base_link",
        target_frame: Optional[str] = "map",
        current_time: Optional["rclpy.time.Time"] = None,
    ):
        """Broadcasts a transform from source to target

        Args:
            transform (Tuple[List[float], List[float]]): a transform consisting of a position and quaternion tuple ([x,y,z], [qx, qy, qz, qw])
            source_frame (Optional[str], optional): the source frame. Defaults to "base_link".
            target_frame (Optional[str], optional): the target frame. Defaults to "map".
            current_time (Optional[&quot;rclpy.time.Time&quot;], optional): current time. Defaults to None.
        """
        t = TransformStamped()

        t.header.stamp = (
            self.node.get_clock().now().to_msg()
            if current_time is None
            else current_time
        )

        translation = transform[0]
        quaternion = transform[1]

        t.header.frame_id = source_frame
        t.child_frame_id = target_frame

        t.transform.translation.x = translation[0]
        t.transform.translation.y = translation[1]
        t.transform.translation.z = translation[2]

        t.transform.rotation.x = quaternion[0]
        t.transform.rotation.y = quaternion[1]
        t.transform.rotation.z = quaternion[2]
        t.transform.rotation.w = quaternion[3]

        self.tf_broadcaster.sendTransform(t)


class TFListener:
    """A transform listener utility class. It can be used query transformations from ROS TF tree."""

    def __init__(self, parent_node: "rclpy.node.Node", buffer_span: float = 0):
        """Constructor to TFListener

        Args:
            parent_node (rclpy.node.Node): parent node to which this TFListener is associated to.
            buffer_span (float, optional): buffer time span in seconds
        """

        self.node = parent_node

        self.tf_buffer = Buffer(Duration(seconds=buffer_span))
        self.tf_listener = TransformListener(self.tf_buffer, self.node)

    def get_transform(
        self,
        base_frame: str,
        source_frame: str,
        time: Optional["rclpy.time.Time"] = None,
        n_trials: int = 5,
    ) -> Tuple[np.ndarray, np.ndarray, "rclpy.time.Time"]:
        """gets transform between base_frame and source_frame at specified time

        Args:
            base_frame (str): base frame name
            source_frame (str): source frame name
            time (Optional[&quot;rclpy.time.Time&quot;], optional):  The time at which to get the transform (0 will get the latest). Defaults to None.

        Returns:
            Tuple[np.ndarray, np.ndarray, &quot;rclpy.time.Time&quot;]: a tuple representing (position, quaternion, time)
        """
        if base_frame == source_frame:
            return (0, 0, 0), (0, 0, 0, 1), self.node.get_clock()
        for i in range(n_trials):
            t = None
            try:
                t = self.tf_buffer.lookup_transform(base_frame, source_frame, time)

                break

            except Exception as e:
                print("Exception type: ", type(e))
                self.node.get_logger().warning(
                    f"Transform could not be queried. Attempt {i}. Warning: {e}"
                )

                if i == (n_trials - 1):
                    self.node.get_logger().error(
                        f"Transform could not be queried. Error: {e}"
                    )
                    return None, None, None

        translation = (
            t.transform.translation.x,
            t.transform.translation.y,
            t.transform.translation.z,
        )
        quaternion = (
            t.transform.rotation.x,
            t.transform.rotation.y,
            t.transform.rotation.z,
            t.transform.rotation.w,
        )

        return (
            np.array(translation).astype(np.float32),
            np.array(quaternion).astype(np.float32),
            time,
        )



import pybullet as p
import numpy as np
from ..math import (
    quaternion2rotation_matrix,
    position_from_matrix,
    quaternion_from_matrix,
)


class BulletSliders:
    def __init__(
        self, phys_id, names, min_vals, max_vals, dft_vals, map_func=lambda _: _
    ):
        """
        Creates a Bullet3 slider GUI components

        Parameters:
            phys_id: <int> physics engine ID
            names: [<str>] labels for each slider
            min_vals: lower limits
            max_vals: upper limits
            dft_vals: default initial values
            map_func: optional map function with signature f(x) -> y used to convert the readings returned by read()
                      default is identity

        """

        assert (
            len(names) == len(min_vals)
            and len(names) == len(max_vals)
            and len(names) == len(dft_vals)
        )

        self.phys_id = phys_id
        self.names = names
        self.min_vals = min_vals
        self.max_vals = max_vals
        self.dft_vals = dft_vals

        self.size = len(self.names)

        self.map_func = map_func

        self.sliders = []

        self.reset()

        self.prev_values = self.read(raw=True)

    def has_changed(self):
        current_values = self.read(raw=True)
        has_changed = not np.isclose(
            np.linalg.norm(self.prev_values - current_values), 0.0
        )

        self.prev_values = current_values

        return has_changed

    def read(self, raw: bool = False):
        values = np.zeros(self.size)
        for i, slider in enumerate(self.sliders):
            values[i] = p.readUserDebugParameter(slider, physicsClientId=self.phys_id)

        return values if raw else self.map_func(values)

    def reset(self):
        self.remove()

        self.sliders = [
            p.addUserDebugParameter(name, ll, ul, dft, physicsClientId=self.phys_id)
            for name, ll, ul, dft in zip(
                self.names, self.min_vals, self.max_vals, self.dft_vals
            )
        ]

    def remove(self):
        for slider in self.sliders:
            p.removeUserDebugItem(slider, physicsClientId=self.phys_id)


class BulletButton(BulletSliders):
    def __init__(self, phys_id: int, name: str):
        """
        Creates a Bullet3 button GUI component

        Args:
            phys_id: physics engine identifier (the id of a BulletWorld)
            name: name and label of this button
        """

        super().__init__(phys_id, [name], [1], [0], [0])

    def was_pressed(self) -> bool:
        """
        Checks if this button was pressed.

        Returns:
            (bool): whether or not this button was pressed since previously checked.
        """

        return self.has_changed()


def add_debug_frame(pos, quat, length=0.2, lw=2, duration=0.2, pcid=0, item_ids=None):
    rot_mat = np.array(p.getMatrixFromQuaternion(quat)).reshape(3, 3).transpose()
    toX = pos + length * rot_mat[0, :]
    toY = pos + length * rot_mat[1, :]
    toZ = pos + length * rot_mat[2, :]
    item_ids = [-1, -1, -1] if not item_ids else item_ids
    item_ids[0] = p.addUserDebugLine(
        pos,
        toX,
        [1, 0, 0],
        lw,
        duration,
        physicsClientId=pcid,
        replaceItemUniqueId=item_ids[0],
    )
    item_ids[1] = p.addUserDebugLine(
        pos,
        toY,
        [0, 1, 0],
        lw,
        duration,
        physicsClientId=pcid,
        replaceItemUniqueId=item_ids[1],
    )
    item_ids[2] = p.addUserDebugLine(
        pos,
        toZ,
        [0, 0, 1],
        lw,
        duration,
        physicsClientId=pcid,
        replaceItemUniqueId=item_ids[2],
    )
    return item_ids


addDebugFrame = add_debug_frame


def add_debug_SE3(
    se3_pose, length: float = 0.1, lw=2, duration: float = 0.0, pcid: int = 0
) -> None:
    """wrapper of add_debug_frame that accepts a SE3 type

    Args:
        se3_pose (_type_): the pose of the frame
        length (float, optional): length of the coordinate frame lines. Defaults to 0.2.
        lw (int, optional): _description_. Defaults to 2.
        duration (float, optional): _description_. Defaults to 0.2.
        pcid (int, optional): pybullet client id. Defaults to 0.
    """
    add_debug_frame(
        se3_pose.translation(),
        se3_pose.quat(),
        length=length,
        lw=lw,
        duration=duration,
        pcid=pcid,
    )


def add_debug_4x4(
    pose_4x4, length: float = 0.1, lw=2, duration: float = 0.0, pcid: int = 0
) -> None:
    """wrapper of add_debug_frame that accepts a SE3 type

    Args:
        pose_4x4 (_type_): the pose of the frame
        length (float, optional): length of the coordinate frame lines. Defaults to 0.2.
        lw (int, optional): _description_. Defaults to 2.
        duration (float, optional): _description_. Defaults to 0.2.
        pcid (int, optional): pybullet client id. Defaults to 0.
    """
    add_debug_frame(
        position_from_matrix(pose_4x4),
        quaternion_from_matrix(pose_4x4),
        length=length,
        lw=lw,
        duration=duration,
        pcid=pcid,
    )


def visualizePcd(pcd, ratio=0.1, duration=10):
    display_freq = int(1 / ratio)
    i = 0
    for p in np.array(pcd.points):
        if (i % display_freq) == 0:
            addDebugFrame(p, [0, 0, 0, 1], length=0.001, duration=10)
        i += 1


visualise_pcd = visualizePcd


def visualize_volume(box, rgb=[1, 0, 0]):
    x_min, x_max = box[0]
    y_min, y_max = box[1]
    z_min, z_max = box[2]

    p1 = [x_min, y_max, z_min]
    p2 = [x_max, y_max, z_min]
    p3 = [x_max, y_min, z_min]
    p4 = [x_min, y_min, z_min]
    p5 = [x_min, y_max, z_max]
    p6 = [x_max, y_max, z_max]
    p7 = [x_max, y_min, z_max]
    p8 = [x_min, y_min, z_max]

    p.addUserDebugLine(p1, p2, rgb)
    p.addUserDebugLine(p1, p5, rgb)
    p.addUserDebugLine(p2, p3, rgb)
    p.addUserDebugLine(p2, p6, rgb)
    p.addUserDebugLine(p3, p4, rgb)
    p.addUserDebugLine(p3, p7, rgb)
    p.addUserDebugLine(p4, p1, rgb)
    p.addUserDebugLine(p4, p8, rgb)
    p.addUserDebugLine(p5, p6, rgb)
    p.addUserDebugLine(p6, p7, rgb)
    p.addUserDebugLine(p7, p8, rgb)
    p.addUserDebugLine(p8, p5, rgb)
    return


def add_text(
    pos, text, text_size=1.2, color=[0, 0, 0], duration=0.0, cid=0, item_id=-1
):
    item_id = p.addUserDebugText(
        text=text,
        textPosition=pos,
        textSize=text_size,
        textColorRGB=color,
        lifeTime=duration,
        replaceItemUniqueId=item_id,
        physicsClientId=cid,
    )

    return item_id


addText = add_text


def draw_frustum(
    cam_pos, cam_ori, hfov, vfov, near, far, ratio, duration=0.0, cid=0, item_ids=None
):
    rot_mat = quaternion2rotation_matrix(cam_ori)
    pos = np.array(cam_pos)

    camRight = rot_mat[:3, 0]
    camUp = rot_mat[:3, 1]
    camForward = -rot_mat[:3, 2]

    pos[0]
    pos[1]
    pos[2]

    nearCenter = pos - camForward * near
    farCenter = pos - camForward * far

    nearHeight = 2 * np.tan(vfov / 2) * near
    farHeight = 2 * np.tan(vfov / 2) * far
    nearWidth = nearHeight * ratio  # 2 * np.tan(hfov / 2) * near
    farWidth = farHeight * ratio  # 2 * np.tan(hfov / 2) * far

    farTopLeft = farCenter + camUp * (farHeight * 0.5) - camRight * (farWidth * 0.5)
    farTopRight = farCenter + camUp * (farHeight * 0.5) + camRight * (farWidth * 0.5)
    farBottomLeft = farCenter - camUp * (farHeight * 0.5) - camRight * (farWidth * 0.5)
    farBottomRight = farCenter - camUp * (farHeight * 0.5) + camRight * (farWidth * 0.5)

    nearTopLeft = nearCenter + camUp * (nearHeight * 0.5) - camRight * (nearWidth * 0.5)
    nearTopRight = (
        nearCenter + camUp * (nearHeight * 0.5) + camRight * (nearWidth * 0.5)
    )
    nearBottomLeft = (
        nearCenter - camUp * (nearHeight * 0.5) - camRight * (nearWidth * 0.5)
    )
    nearBottomRight = (
        nearCenter - camUp * (nearHeight * 0.5) + camRight * (nearWidth * 0.5)
    )

    if item_ids is None:
        item_ids = [-1] * 16

    lw = 2
    # Near pyramid
    item_ids[0] = p.addUserDebugLine(
        pos,
        nearTopLeft,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[0],
        physicsClientId=cid,
    )
    item_ids[1] = p.addUserDebugLine(
        pos,
        nearTopRight,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[1],
        physicsClientId=cid,
    )
    item_ids[2] = p.addUserDebugLine(
        pos,
        nearBottomLeft,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[2],
        physicsClientId=cid,
    )
    item_ids[3] = p.addUserDebugLine(
        pos,
        nearBottomRight,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[3],
        physicsClientId=cid,
    )

    item_ids[4] = p.addUserDebugLine(
        pos,
        farTopLeft,
        [1, 0, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[4],
        physicsClientId=cid,
    )
    item_ids[5] = p.addUserDebugLine(
        pos,
        farTopRight,
        [1, 0, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[5],
        physicsClientId=cid,
    )
    item_ids[6] = p.addUserDebugLine(
        pos,
        farBottomLeft,
        [1, 0, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[6],
        physicsClientId=cid,
    )
    item_ids[7] = p.addUserDebugLine(
        pos,
        farBottomRight,
        [1, 0, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[7],
        physicsClientId=cid,
    )

    item_ids[8] = p.addUserDebugLine(
        nearTopLeft,
        nearTopRight,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[8],
        physicsClientId=cid,
    )

    item_ids[9] = p.addUserDebugLine(
        nearTopRight,
        nearBottomRight,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[9],
        physicsClientId=cid,
    )

    item_ids[10] = p.addUserDebugLine(
        nearBottomRight,
        nearBottomLeft,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[10],
        physicsClientId=cid,
    )

    item_ids[11] = p.addUserDebugLine(
        nearBottomLeft,
        nearTopLeft,
        [0, 1, 0],
        lw,
        duration,
        replaceItemUniqueId=item_ids[11],
        physicsClientId=cid,
    )

    item_ids[12] = p.addUserDebugLine(
        farTopLeft,
        farTopRight,
        [0, 0, 1],
        lw,
        duration,
        replaceItemUniqueId=item_ids[12],
        physicsClientId=cid,
    )

    item_ids[13] = p.addUserDebugLine(
        farTopRight,
        farBottomRight,
        [0, 0, 1],
        lw,
        duration,
        replaceItemUniqueId=item_ids[13],
        physicsClientId=cid,
    )

    item_ids[14] = p.addUserDebugLine(
        farBottomRight,
        farBottomLeft,
        [0, 0, 1],
        lw,
        duration,
        replaceItemUniqueId=item_ids[14],
        physicsClientId=cid,
    )

    item_ids[15] = p.addUserDebugLine(
        farBottomLeft,
        farTopLeft,
        [0, 0, 1],
        lw,
        duration,
        replaceItemUniqueId=item_ids[15],
        physicsClientId=cid,
    )
    return item_ids


def draw_aabb(aabb, duration=0, phys_id=0):
    aabbMin = aabb[0]
    aabbMax = aabb[1]
    f = [aabbMin[0], aabbMin[1], aabbMin[2]]
    t = [aabbMax[0], aabbMin[1], aabbMin[2]]
    p.addUserDebugLine(f, t, [1, 0, 0], lifeTime=duration, physicsClientId=phys_id)
    f = [aabbMin[0], aabbMin[1], aabbMin[2]]
    t = [aabbMin[0], aabbMax[1], aabbMin[2]]
    p.addUserDebugLine(f, t, [0, 1, 0], lifeTime=duration, physicsClientId=phys_id)
    f = [aabbMin[0], aabbMin[1], aabbMin[2]]
    t = [aabbMin[0], aabbMin[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [0, 0, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMin[0], aabbMin[1], aabbMax[2]]
    t = [aabbMin[0], aabbMax[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMin[0], aabbMin[1], aabbMax[2]]
    t = [aabbMax[0], aabbMin[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMax[0], aabbMin[1], aabbMin[2]]
    t = [aabbMax[0], aabbMin[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMax[0], aabbMin[1], aabbMin[2]]
    t = [aabbMax[0], aabbMax[1], aabbMin[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMax[0], aabbMax[1], aabbMin[2]]
    t = [aabbMin[0], aabbMax[1], aabbMin[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMin[0], aabbMax[1], aabbMin[2]]
    t = [aabbMin[0], aabbMax[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)

    f = [aabbMax[0], aabbMax[1], aabbMax[2]]
    t = [aabbMin[0], aabbMax[1], aabbMax[2]]
    p.addUserDebugLine(
        f, t, [1.0, 0.5, 0.5], lifeTime=duration, physicsClientId=phys_id
    )
    f = [aabbMax[0], aabbMax[1], aabbMax[2]]
    t = [aabbMax[0], aabbMin[1], aabbMax[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)
    f = [aabbMax[0], aabbMax[1], aabbMax[2]]
    t = [aabbMax[0], aabbMax[1], aabbMin[2]]
    p.addUserDebugLine(f, t, [1, 1, 1], lifeTime=duration, physicsClientId=phys_id)


drawAABB = draw_aabb


def draw_cloud(
    pcd: "open3d.geometry.PointCloud",
    subsample_size=5000,
    pcd_debug_item: int = -1,
    phys_id: int = 0,
):
    all_points = np.asarray(pcd.points)
    ids = np.random.choice(np.arange(0, all_points.shape[0]), subsample_size)

    colors = np.zeros_like(pcd.colors)[ids, :]

    points = np.asarray(pcd.points)[ids, :]

    p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0, physicsClientId=phys_id)

    if pcd_debug_item >= 0:
        p.addUserDebugPoints(
            points,
            colors,
            lifeTime=0,
            pointSize=2,
            replaceItemUniqueId=pcd_debug_item,
            physicsClientId=phys_id,
        )
    else:
        pcd_debug_item = p.addUserDebugPoints(points, colors, lifeTime=0, pointSize=2)

    p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)

    return pcd_debug_item



from .depth_proc_launch import depth_proc_launch
from .io import (
    xacro_to_urdf_path,
    xacro_to_urdf_string,
    is_sim,
    is_real_robot_connected,
)
from .joint_state_tools import JointStatePub, JointStateSub, LocalJointStatePub
from .ros_camera import ROSCamera
from .ros_robot_arm import ROSRobotArm
from .tf_tools import TFListener, TFPublisher
from .trajectory_action_server import TrajectoryActionServer

__all__ = [
    "xacro_to_urdf_path",
    "xacro_to_urdf_string",
    "is_sim",
    "is_real_robot_connected",
    "JointStatePub",
    "LocalJointStatePub",
    "ROSCamera",
    "ROSRobotArm",
    "TFListener",
    "TFPublisher",
    "TrajectoryActionServer",
]



#!/usr/bin/env python3

import os
import numpy as np

try:
    import rclpy
    from rclpy.node import Node

except ImportError:
    raise RuntimeError("Unable to run example without ros2 installed.")

import grip
import grip.ros
import pybullet_data as pd


class ROSCameraExample(Node):
    def __init__(self, fps=10):
        super().__init__("rgbd_camera_example_node")

        # Creates bullet world with GUI
        self.world = grip.robot.BulletWorld(phys_opt="gui")

        # Creates ground plane object
        self.plane = grip.robot.BulletObject(
            urdf_file=os.path.join(pd.getDataPath(), "plane.urdf"), position=[0, 0, 0]
        )

        # Creates ROSCamera object
        self.camera = grip.ros.ROSCamera(
            cid=self.world.id,
            position=np.array([0, 0, 1.5]),
            orientation=grip.math.rpy2quaternion([0.0, np.pi, 0.0]),
            parent_node=self,
        )

        self.idx = 0
        self.angs = np.linspace(np.pi, 1.5 * np.pi, 100)

        self.timer = self.create_timer(1 / fps, self.on_update)

    def on_update(self):
        rgb, depth = self.camera.obs()

        ang = self.angs[self.idx]

        self.camera.update_view_from_pose(
            np.array([0, 0, 1.5 + np.sin(ang * 0.1)]),
            grip.math.rpy2quaternion([0.0, ang, 0.0]),
        )

        self.camera.draw_frustum(
            0.0,
        )

        self.idx = (self.idx + 1) % len(self.angs)


def main(args=None):
    rclpy.init(args=args)

    ros_camera_example = ROSCameraExample()

    rclpy.spin(ros_camera_example)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    ros_camera_example.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



import numpy as np
import cv2
from PIL import Image


def visualize_depth(depth_in, title, min_depth=None, max_depth=None, wait_time=0):
    if min_depth is None:
        min_depth = np.nanmin(depth_in)
    if max_depth is None:
        max_depth = np.nanmax(depth_in)
    depth_vis = np.clip(depth_in, min_depth, max_depth)
    depth_vis = depth_vis / max_depth
    # depth_vis = (depth_vis - min_depth) / (max_depth - min_depth)
    depth_vis *= 255.0
    depth_vis = depth_vis.astype(np.uint8)
    cv2.imshow(title, depth_vis)
    k = cv2.waitKey(wait_time)
    if wait_time == 0:
        cv2.destroyWindow(title)
    return k


def resize_numpy_image(image, size, resize_type="antialias"):
    assert resize_type in [
        "nearest",
        "antialias",
        "bicubic",
        "bilinear",
    ], "Resize type invalid"
    assert image.dtype == "uint8", "Input image array must be in np.uint8 type"
    image = image.astype(np.uint8)
    if resize_type == "nearest":  # for segmentation mask resize
        image = image.resize(size, Image.NEAREST)
    elif resize_type == "antialias":
        image = image.resize(size, Image.ANTIALIAS)
    elif resize_type == "bicubic":
        image = image.resize(size, Image.BICUBIC)
    elif resize_type == "bilinear":
        image = image.resize(size, Image.BILINEAR)
    image = np.array(image)
    return image


def pcd_to_image(obj_pcd, cam_intrinsic, cam_extrinsic=None):
    w = cam_intrinsic.width
    h = cam_intrinsic.height
    fx, fy = cam_intrinsic.get_focal_length()
    cx, cy = cam_intrinsic.get_principal_point()
    n_points = obj_pcd.shape[0]

    projected_image = np.zeros((h, w))
    projected_image[:] = np.NaN

    if n_points > 0:
        if cam_extrinsic is not None:
            cam_xyz = np.matmul(
                cam_extrinsic,
                np.expand_dims(np.c_[obj_pcd, np.ones(n_points)], axis=-1),
            )
            cam_xyz = cam_xyz.squeeze(-1)[:, :-1]
        else:
            cam_xyz = obj_pcd.copy()

        cam_xyz[:, 0] = np.round((cam_xyz[:, 0] * fx / cam_xyz[:, -1]) + cx)
        cam_xyz[:, 1] = np.round((cam_xyz[:, 1] * fy / cam_xyz[:, -1]) + cy)

        in_range_mask = np.logical_and(
            np.logical_and(
                np.logical_and(cam_xyz[:, 0] < w, cam_xyz[:, 1] < h), cam_xyz[:, 0] >= 0
            ),
            cam_xyz[:, 1] >= 0,
        )
        in_range_cam_xyz = cam_xyz[in_range_mask]
        projected_image[
            in_range_cam_xyz[:, 1].astype(int), in_range_cam_xyz[:, 0].astype(int)
        ] = in_range_cam_xyz[:, -1]

    return projected_image
